/*
	This file contains all routines to utilize the AES.CFB mode of decryption.
	(CFB = Cipher Feedback)

    Status as of 2009.03.22:
        This program produces the same output as the DMB Bootloader program

    Note:   All two-dimensional arrays were converted to linear arrays due to
            problems with the CCS compiler not able to successfully build
            functional code passing such arrays to functions.



	Encryption:

	1st Block:

	IV + KEY = Block Cypher Encryption (BCE)
	BCE XOR PLAINTEXT = CYPHERTEXT#1

	2nd Block:

	CYPHERTEXT#1 + KEY = BCE
	BCE XOR PLAINTEXT = CYPHERTEXT#2

	repeat until done...

	IV and KEY are 16 bytes long

	Decryption:

	1st Block:

	IV + KEY = Block Cypher Encryption (BCE#1)
	BCE#1 XOR CYPHERTEXT#1 = PLAINTEXT#1

	2nd Block:

	CYPHERTEXT#1 + KEY = BCE#2

	BCE#2 XOR CYPHERTEXT#2 = PLAINTEXT#2

	repeat until done...



	CFB Decrypt Function:
	=================================================================================

	                          +----------------------+
	                          |                      |
	(Encryptded) Data Array-->|        CFB           |----> Data Array (Decrypted)
	                          |   Decrypt function   |
	KEY---------------------->|                      |
	INITIAL VECTOR----------->|                      |----+
	                   ^      |                      |    |
                       |      +----------------------+    |
                       |                                  |
                       +------New initial vector----------+

	Note: Function uses a pointer to read from and write to the data array containg initially
	      the encrypted data. It is replaced in 16 byte blocks by decrypted data until done.



	Source Packet HEX Dump:
	aa 5c 72 6e 9c 09 be bb d0 2b 96 6f e7 52 70 2b
	fd 61 51 67 aa 74 65 0b 36 8a 30 f0 21 5d 27 cb

	aa 8e c4 05 a1 56 51 f8 b2 7c bb 96 dd e3 6b 6b
	52 ef db 77 a2 e2 d4 ec 16 af b4 ba fa 04 9e 2c

	ef 9b 72 38 af db a1 9c 7c 34 66 5e 41 4c 1a e6
	87 8b 35 9e 56 cf 07 6e 47 3b 8a 2e d6 1d f9 c9

	4b 22 9b 24 c4 64 af 93 c6 a0 b4 05 15 ac 49 3c
	40 b1 c0 09 ff b9 91 35 df ca 62 bc 04 24 8b 0e

	Decrypted Packet:
	4e 5c 00 c8 09 00 01 00 00 00 00 00 00 00 00 00
	00 00 cc 45 04 00 00 00 cc 45 04 00 00 00 cc 45

	04 00 00 00 cc 45 04 00 00 00 cc 45 04 00 00 00
	cc 45 04 00 00 00 cc 45 04 00 00 00 cc 45 04 00

	00 00 cc 45 04 00 00 00 cc 45 04 00 00 00 56 45
	04 00 00 00 cc 45 04 00 00 00 cc 45 04 00 00 00

	cc 45 04 00 00 00 cc 45 04 00 00 00 cc 45 04 00
	00 00 cc 45 04 00 00 00 cc 45 04 00 00 00 cc 45


OK	4e 5c 00 c8 09 00 01 00 00 00 00 00 00 00 00 00

OK	00 00 cc 45 04 00 00 00 cc 45 04 00 00 00 cc 45

	53 fd d7 05 b5 1c bb cb 5e 0e 73 49 25 9d 6c 2d

OK	cc 45 04 00 00 00 cc 45 04 00 00 00 cc 45 04 00

	71 54 70 e3 86 29 e3 cf 81 b5 c2 83 b3 a8 72 1a

OK	04 00 00 00 cc 45 04 00 00 00 cc 45 04 00 00 00

	c0 cf 63 6b c9 5f 73 a6 98 47 80 26 42 6f 5b 6f

OK	00 00 cc 45 04 00 00 00 cc 45 04 00 00 00 cc 45





*/

// Setup for 128 BIT CFB
#define KEY_ROW 4
#define KEY_COLUMN 4
#define AES_BLOCK_SIZE 16
#define KEY_LENGTH 128
#define ROUND_KEY_COLUMN (4*(10+(KEY_LENGTH-128)/32)+KEY_LENGTH/32)//=44
#define ARRAY_BLOCK_SIZE (KEY_ROW * KEY_COLUMN)//=16
#define ROUNDKEY_SIZE (KEY_ROW * ROUND_KEY_COLUMN)// 4x44=176 bytes

#define TEST			// Enable main() test code
#define ADD_ENCRYPT	// If you want encryption
//#define ADD_ENCRYPT_TEST




uint16_t aes_cfb_encrypt(uint8_t *data, uint8_t *iv, uint16_t size);
uint16_t aes_cfb_decrypt(uint8_t *data, uint8_t *iv, uint16_t size);

uint16_t encrypt(uint8_t *data_in);// ROM CYPHER
void sub_bytes(uint8_t *data);
void shift_rows(uint8_t *data);
void mix_columns(uint8_t *data);
void shift_row(uint32_t *data, uint16_t row, uint16_t shift);
void add_round_key(uint8_t *data, uint8_t *key, uint8_t key_idx);
void count_key_schedule(uint8_t *round_key);// ROM CYPHER

// 128 bit key (conerted to linear array from 4x4 original format
//uint8_t key[16] =
uint8_t const fl_key[16] =
{
    0x24, 0x43, 0xeb, 0x9c,
    0x57, 0xda, 0xe1, 0xe2,
    0x45, 0x86, 0x6a, 0xb3,
    0x5c, 0xf6, 0x43, 0x06
};

// init Vectors key
//uint8_t iv[16] = {
uint8_t const fl_init_vectors[16] = {
    0x29, 0x87, 0xA4, 0xFB,
    0xBA, 0xF0, 0x4F, 0xA8,
    0x34, 0x30, 0xBC, 0x38,
    0xBD, 0x77, 0x5B, 0xB1
  };



/**
 * Rijndael S-box.
 */
uint8_t const s_box[16][16] = {
  {0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76},
  {0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0},
  {0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15},
  {0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75},
  {0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84},
  {0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf},
  {0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8},
  {0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2},
  {0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73},
  {0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb},
  {0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79},
  {0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08},
  {0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a},
  {0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e},
  {0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf},
  {0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16}
};



/**
 * Round constant word array.
 */
uint8_t const rcon[255] = {
  0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b,
  0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63,
  0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5,
  0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f, 0x25,
  0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 0x74, 0xe8,
  0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
  0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc,
  0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef,
  0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f,
  0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 0x74,
  0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40,
  0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e,
  0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa,
  0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2,
  0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a,
  0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20,
  0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f,
  0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d,
  0xfa, 0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61,
  0xc2, 0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d,
  0x3a, 0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10,
  0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a,
  0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3,
  0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd,
  0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83,
  0x1d, 0x3a, 0x74, 0xe8, 0xcb
};


extern enum {
  AES_ERR_NOERR   = 0,  // no error
  AES_ERR_INPUT   = 1,  // problem with input file
  AES_ERR_OUTPUT  = 2,  // problem with output file
  AES_ERR_MALLOC  = 3,  // memory can not be allocated
  AES_ERR_KEYSIZE = 4,  // an inappropriate key size (only 128, 192 i 256 are allowed)
  AES_ERR_FORMAT  = 5   // an innapropriate file format
};// aes_error;


/**
 * \brief Encrypts file_in with AES CFB algorithm using Cypher Key and saves it to file_out.
 *        If initialization vector is not set on arguments list, it will be written to the output file.
 * \param file_in     [in] Name of a file to be encrypted.
 * \param file_out    [in] Name of an encrypted file.
 * \param key         [in] Cypher Key.
 * \param key_size    [in] Cypher Key length.
 * \param init_vector [in] Pointer to an initialization vector. If set to NULL, iv will randomized.
 * \return 0 on success, error code otherwise.
 */
#ifdef ADD_ENCRYPT
uint16_t aes_cfb_encrypt(uint8_t *data, uint8_t *iv, uint16_t size)
{
	uint16_t  length;
	uint16_t  i,j=0;
	uint16_t  errCode = AES_ERR_NOERR;   	// error code (if any)
	uint8_t iv_last[AES_BLOCK_SIZE];         	// initialization vector
	unsigned char text[AES_BLOCK_SIZE];  	// piece of text to be encryptd

	// 5. until eof read file by AES_BLOCK_SIZE bytes pieces
	do
	{
		if (size >= AES_BLOCK_SIZE )
		{
			length = AES_BLOCK_SIZE;
			size -= AES_BLOCK_SIZE;
		}
		else
		{
			length = size;
			size = 0;
		}

		// 5.1 encrypt initialization vector

		// The initial Vector table must be preserved bevor exiting function
		memcpy(iv_last, iv, AES_BLOCK_SIZE);

		errCode = encrypt(iv);

		if (AES_ERR_NOERR != errCode)
			return (errCode);

		// 5.2 read next AES_BLOCK_SIZE bytes
		memcpy(text, data, AES_BLOCK_SIZE);

		// 5.3 xor input plaintext with initialization vector
		if (errCode == 0)
		{
			if (length == AES_BLOCK_SIZE)
			{
				// Encrypt source data block
				for (i = 0; i < length; i++)
				{
					iv[i] ^= text[i];
				}
			}
			else
			{
				memcpy(iv, iv_last, AES_BLOCK_SIZE);
				break;
			}
			// 5.6 write data back
			memcpy(data, iv, AES_BLOCK_SIZE);
			data = data + length;
		}

	} while (length == AES_BLOCK_SIZE);

	return (errCode);
}
#endif



/**
 * \brief Decrypts data with AES CFB algorithm using Cypher Key.
 * \param data        [in out] data to be decrypted.
 * \param key         [in] Cypher Key.
 * \param key_size    [in] Cypher Key length.
 * \param init_vector [in out] Pointer to an initialization vector. If set to NULL, iv will be read from an input file.
 * \return 0 on success, error code otherwise.
 */
//  data, size , key, KEY_LENGTH, iv

uint16_t aes_cfb_decrypt(uint8_t *data, uint8_t *iv, uint16_t size )
{

	uint16_t  length;
	uint16_t  i;
	uint16_t  errCode = AES_ERR_NOERR;   // error code (if any)
	//uint8_t iv[AES_BLOCK_SIZE];         // initialization vector
	uint8_t next_iv[AES_BLOCK_SIZE];    // next initialization vector

	output_high(PIN_D3);

	if (errCode == 0)
	{
		do
		{
			// 5.1 ENCRYPT initialization vector
			errCode = encrypt(iv);    // uses: iv, key

			// 5.1 read next AES_BLOCK_SIZE bytes
			if (size >= AES_BLOCK_SIZE )
			{
				length = AES_BLOCK_SIZE;
				size -= AES_BLOCK_SIZE;
			}
			else
			{
				length = size;
				size = 0;
			}

			if (length == AES_BLOCK_SIZE)
			{
				memcpy(next_iv, data, AES_BLOCK_SIZE);

				if (errCode == 0)
				{
				  	// 5.3 xor input plaintext with initialization vector
					for (i = 0; i < AES_BLOCK_SIZE; i++)
					{
						data[i] ^= iv[i];
					}
					// transfer Cyphertext to IV for next decryption pass
					memcpy(iv, next_iv, AES_BLOCK_SIZE);
					data = data + length;
				}
			}
			else
			{
				// Copy Cypher to be used decrypt() when called again
				memcpy(iv, next_iv, AES_BLOCK_SIZE);
				break;
			}
		}while ((errCode == 0) && (length == AES_BLOCK_SIZE));
	}

	output_low(PIN_D3);

	return (errCode);
}



/**
 * \brief Transforms the data array using a non-linear byte substitution array.
 * \param data [in/out] Data to be substitutet.
 */
void sub_bytes(uint8_t *data)
{
	uint16_t i, j, k;
	uint8_t c;

	// 1. for each element in data matrix
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < 4; j++)
		{
			k = j * 4;
            // 1.1 replace it by the proper value from s-box
			c = data[k + i];
			data[k + i] = s_box[(c >> 4) & 0x0F][c & 0x0F];
		}
	}
}



/**
 * \brief Cyclically shifts right the last three rows of data array.
 * \param data [in/out] Matrix with rows to be shifted.
 */
void shift_rows(char *data)
{
	// 1. shift 2nd row by 1
	shift_row(data, 1, 1);
	// 2. shift 3rd row by 2
	shift_row(data, 2, 2);
	// 3. shift 4th row by 3
	shift_row(data 3, 3);
}



/**
 * \brief Cyclically shifts right a determined row by shift.
 * \param data  [in/out] Matrix to have row shifted.
 * \param row   [in]     Number of row to be shifted.
 * \param shift [in]     The size of what the row should be shifted.
 */
void shift_row(uint32_t *data, uint16_t row, uint16_t cols)
{
	uint8_t i, temp;
	temp = cols  * 8;
 	for(i=0;i<(temp);i++)
	{
		rotate_right( &data[row], 4);
	}
}



/**
 * \brief Mixes data of columns to produce new column.
 * \param data [in/out] Matrix to have mixed columns.
 */
void mix_columns(uint8_t *data)
{
	uint16_t c, r;
	uint8_t base[4];
	uint8_t base2[4];  // base2[i] = 2 * base[i];

	// 1. for each column in data matrix
	for (c = 0; c < 4; c++)
	{
		// 1.1 multiply it by MDS matrix (this is multiplication operation in Galois Field)
		for(r = 0; r < 4; r++)
		{
			base[r] = data [(r * 4) + c];
			base2[r] = base[r] << 1;  // * 2
			if ((base[r] & 0x80) == 0x80)
				base2[r] ^= 0x1b;
		}

		data[     c] = base2[0] ^ base[3] ^ base[2] ^ base2[1] ^ base[1];
		data[ 4 + c] = base2[1] ^ base[0] ^ base[3] ^ base2[2] ^ base[2];
		data[ 8 + c] = base2[2] ^ base[1] ^ base[0] ^ base2[3] ^ base[3];
		data[12 + c] = base2[3] ^ base[2] ^ base[1] ^ base2[0] ^ base[0];
	}
}



/**
 * \brief Encrypts input data.
 * \param data_in    [in/out] Data to be encrypted.
 * \param cypher_key [in]     Cypher Key.
 * \param key_size   [in]     Cypher Key length (in bits).
 * \return 0 on success, error code otherwise.
 */
uint16_t encrypt(uint8_t *data_in)// uses iv, key as input from ROM
{

  	uint16_t counter, i;
	uint8_t round_key[ROUNDKEY_SIZE];
	uint8_t data[16];
	uint8_t buf[4][4];

	counter = 10+(KEY_LENGTH-128)/32;

  	// 2. convert data to a useful format (read by columns)
  	for (i = 0; i < AES_BLOCK_SIZE; i++) {
    	buf[i%4][i/4] = data_in[i];
  	}

	memcpy(data, buf, AES_BLOCK_SIZE);
  	counter -= 1;

  	// 3. count key schedule
  	count_key_schedule(round_key);	// cypher from ROM

  	// 4. Add Round Key
    add_round_key(data, round_key, 0);

  	// 5. do it (10+(key_size-128)/32)-1 times, = 9
  	i = 0;
  	while (i < counter)
  	{
  		// 5.1. Sub Bytes
        sub_bytes(data);

  		// 5.2. Shift Rows
        shift_rows(data);

  		// 5.3. Mix Columns
        mix_columns(data);

  		// 5.4. Add Round Key
        add_round_key(data, round_key, 4*(++i));

  	}
	// 6. Final Round
	// 6.1 Sub Bytes
    sub_bytes(data);

	// 6.2 Shift Rows
    shift_rows(data);

	// 6.3 Add Round Key
    add_round_key(data, round_key, 4*(++i));

	// 7. copy data back
	memcpy(buf, data, AES_BLOCK_SIZE);

  	// 7.1 convert data to a useful format - convert columns to rows
  	for (i = 0; i < AES_BLOCK_SIZE; i++) {
    	data_in[i] = buf[i%4][i/4];
  	}

	return AES_ERR_NOERR;
}



/**
 * \brief Adds the Round Key to the data using an XOR operation.
 * \param data    [in/out]  Data to be {enc,dec}rypted.
 * \param key     [in]      Round Key to be added.
 * \param key_idx [in]      Index of the key's column, where should adding start.
 */
void add_round_key(uint8_t *data, uint8_t *key, uint8_t key_idx)
{
	uint16_t c, r;
/*
	XOR round key to data
	The original row/column array are substituted by linear arrays
*/
	// 1. for each column in data matrix
	for (c = 0; c < 4; c++)
	{
		// 1.1 add proper column from Round Key
		for (r = 0; r < 4; r++)
		{
            data[(r*4) + c] ^= key[(r * ROUND_KEY_COLUMN) + key_idx + c];
		}
	}
}



/**
 * \brief Calculates Round Key.
 * \param data       [in]  Data to be {enc,dec}rypted.
 * \param round_key  [out] Round key to be added.
 * \param key_length [in]  Key length.
 */
void count_key_schedule( uint8_t *round_key)// ROM Cypher
{
    uint16_t i, j, k, kk;
    uint8_t  c;
    uint16_t counter;
    uint16_t nk;

    nk      = KEY_LENGTH/32;//=4
    counter = 10 + nk - 4;//=10

    // 1. 'copy' a cypher key to a round key (this is not simple copy)
    for (i = 0; i < nk; i++) 	// COL INDX
    {
		for (j = 0; j < 4; j++)	// ROW INDX
		{
			// calculate row offset index
			k = j * ROUND_KEY_COLUMN;
            round_key[ k + i ] = fl_key[ (j*4) + i ];// ROM
		}
  	}

    // 2. for i from 3 to 4*counter+nk-1 do the following:
    for (i = nk; i < (4 * (counter + 1)); i++)
    {
		// for column W_i from cypher_key do:
		if (0 == (i % nk))
		{
			// 1.1. copy W_i-1 column
			//      and rotate word from this column
			// calculate row offset index
			k = 3 * ROUND_KEY_COLUMN;
			round_key[k + i] = round_key[ i - 1];

			for (j = 1; j < 4; j++)
			{
				// calculate row offset index
				k = (j-1) * ROUND_KEY_COLUMN;
				kk = j * ROUND_KEY_COLUMN;
				round_key[k + i] = round_key[ kk + i - 1];
			}

			for (j = 0; j < 4; j++)
			{
				// calculate row offset index
				k = j * ROUND_KEY_COLUMN;
				// 1.2. subbytes on this rotated column
				c = round_key[k + i];
				round_key[k + i]  = s_box[(c >> 4) & 0x0F][c & 0x0F];
				// 1.3. add (do a bitwise OR) the W_i from round_key column
				round_key[k + i] ^= round_key[k + i - nk];
			}

			// 1.4. add (do a bitwise OR) the i-th element of Rcon array to the W_i[0]
			k=0;
			round_key[k + i] ^= rcon[i / nk];
		}
		else if ((6 < nk) && (4 == i % nk))
		{
			for (j = 0; j < 4; j++)
			{
				// calculate row offset index
				k = j * ROUND_KEY_COLUMN;
				// 1.1. subbytes on this column
				c = round_key[k + i - 1];
				round_key[k + i]  = s_box[(c >> 4) & 0x0F][c & 0x0F];
				// 1.3. add (do a bitwise OR) the W_i from round_key column
				round_key[k + i] ^= round_key[k + i - nk];
			}
      	}
		else
		{
			// 1.1. copy W_i-1 column (if (i-1 < 0) get column from round_key) to W_i
			// and
			// 1.2. add (do a bitwise OR) the W_i from round_key column
			for (j = 0; j < 4; j++)
			{
			  // calculate row offset index
			  k = j * ROUND_KEY_COLUMN;
			  round_key[k + i]  = round_key[k + i - 1];
			  round_key[k + i] ^= round_key[k + i - nk];
			}
		}
    }

}

