/*
    SPM Bootloader AES decryption Test Program and Reference Design
    ===============================================================

    This program decrypts the 128 Test Pattern using the
    CFB Mode of operation.

    The decryption requires the Cypher Key and the initialization vectors
    (these are organized as 128 Bits of 4x4 Bytes)

    Date:   2009/03/22
    Who:    GJO
    Compiler used:  CCS Ver 4.087
    Target uC:      PIC33FJ256MC710

    Performance at 40MHZ is 5.702KB/s, or 0.17536ms/Byte or 22.44ms/128 bytes

    History:
    =================================================================================

    09/04/15	GJO		Added multiple pass capability so that encryption and
    					decryption is possible in a distributed fashion by storing
    					the intermediate initial vector data in a global array.
    					This is required in the boot loader.



*/
#include <33FJ256MC510.h>
#device adc=10

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#fuses NODEBUG, NOWRTB, NOWRTSS
#fuses NOWRT, NOPROTECT, NORSS,

#fuses XT, PR_PLL, NOCOE, NOWDT, NOIESO   // XT Oscillator with PLL enabled

#use delay( clock=40000000) //Set clock as required
#use rs232( baud=57600, rcv=PIN_F4, xmit=PIN_F5, UART2, parity=N, BITS=8, STOP=1 )

#word PLLFBD    = 0x0746
#word CLKDIV    = 0x0744

#define PORT_A_MASK 0x0000

#include "SPM_BL_types.h"

#define NUM_BYTES 128	// Input data size must be multiples of 16 bytes

#define SPM_BIN 0
#define QUCK_BROWN_FOX 1

#define IN_DATA SPM_BIN

#if IN_DATA == SPM_BIN

uint8_t data_in[NUM_BYTES] =
{
0xAA, 0x5C, 0x72, 0x6E, 0x9C, 0x09, 0xBE, 0xBB,
0xD0, 0x2B,	0x96, 0x6F,	0xE7, 0x52,	0x70, 0x2B,
0xFD, 0x61,	0x51, 0x67,	0xAA, 0x74,	0x65, 0x0B,
0x36, 0x8A,	0x30, 0xF0,	0x21, 0x5D,	0x27, 0xCB,
0xAA, 0x8E,	0xC4, 0x05,	0xA1, 0x56,	0x51, 0xF8,
0xB2, 0x7C,	0xBB, 0x96,	0xDD, 0xE3,	0x6B, 0x6B,
0x52, 0xEF,	0xDB, 0x77,	0xA2, 0xE2,	0xD4, 0xEC,
0x16, 0xAF,	0xB4, 0xBA,	0xFA, 0x04,	0x9E, 0x2C,
0xEF, 0x9B,	0x72, 0x38,	0xAF, 0xDB,	0xA1, 0x9C,
0x7C, 0x34,	0x66, 0x5E,	0x41, 0x4C,	0x1A, 0xE6,
0x87, 0x8B,	0x35, 0x9E,	0x56, 0xCF,	0x07, 0x6E,
0x47, 0x3B,	0x8A, 0x2E,	0xD6, 0x1D,	0xF9, 0xC9,
0x4B, 0x22,	0x9B, 0x24,	0xC4, 0x64,	0xAF, 0x93,
0xC6, 0xA0,	0xB4, 0x05,	0x15, 0xAC,	0x49, 0x3C,
0x40, 0xB1,	0xC0, 0x09,	0xFF, 0xB9,	0x91, 0x35,
0xDF, 0xCA,	0x62, 0xBC,	0x04, 0x24,	0x8B, 0x0E
};
#else


uint8_t data_in[NUM_BYTES] = {"The quick brown fox jumps over the lazy dog. The quick brown fox jumps over the lazy dog.The quick brown fox jumps over the laz"};

#endif


#include "AES_ALL.c"

void main(void)
{

	uint16_t i,j=0, byte_block;

	uint8_t init_vectors[AES_BLOCK_SIZE];

    output_a(0x0000);
//    set_tris_a(PORT_A_MASK);
//    set_uart_speed(57600);
    PLLFBD    = 0x26;
    CLKDIV    &= 0xFFFC;


	// Copy initial vectors from FLASH to RAM because FLASH canot be accessed
	// with pointers
	memcpy(init_vectors, fl_init_vectors, AES_BLOCK_SIZE);


#ifdef TEST
	// Wait for PLL to stabilize before the UART can be used
    delay_ms(100);

	printf("\nAES 128 BIT CFB DECRYPTION TEST PROGRAM\n");

	printf("\nSource Packet HEX Dump:\n\n");

	j=0;
	for(i=0;i<NUM_BYTES;i++){
		printf("%02x ", data_in[i]);
		if(++j==16){
			j=0;
			printf("\n");
		}
	}

 #ifdef ADD_ENCRYPT_TEST
	printf("\nPacket Characters: \n\n");
	j=0;
	for(i=0;i<NUM_BYTES;i++)
	{
		if((data_in[i] >= 0x20) && (data_in[i] <= 0x7F))
		{
			printf("%c", data_in[i]);
		}
		else
		{
			printf("*");
		}
		if(++j==64){
			j=0;
			printf("\n");
		}
	}

	// Uses existing key and initial vectors
	// Adapted from DCAPP (Smartpack.exe program)

	memcpy(init_vectors, fl_init_vectors, AES_BLOCK_SIZE);
	aes_cfb_encrypt(data_in, init_vectors, NUM_BYTES);	// ROM KEY and INIT Vectors

	printf("\nEncrypted Packet: \n\n");
	j=0;
	for(i=0;i<NUM_BYTES;i++){
		printf("%02x ", data_in[i]);
		if(++j==16){
			j=0;
			printf("\n");
		}
	}

 	printf("\nEncrypted Packet Characters: \n\n");
 	j=0;
 	for(i=0;i<NUM_BYTES;i++)
 	{
 		if((data_in[i] >= 0x20) && (data_in[i] <= 0x7F))
 		{
 			printf("%c", data_in[i]);
 		}
 		else
 		{
 			printf("*");
 		}
 		if(++j==64){
 			j=0;
 			printf("\n");
 		}
	}

 #endif

#endif


#ifdef LOOPTEST
	i=1872;// is aequivalent to 240KB of PIC FLASH ROM
	printf("\nStart decryption of %u blocks: ", i);
	while(i--){
	    aes_cfb_decrypt(data_in, init_vectors, NUM_BYTES,);
		j=0;
		if(++j==64){
			j=0;
			printf("\n");
		}
	}
	printf("\nEnd decryption: ");
#else
/*
	This is a test of chained operation of decrypt function to
	ensure correct operation with the bootloader

	The purpose of this test is to verify that input data can be
	processed in continuous external blocks by using the init_vector
	passed back again as the input for the subsequent use of decrypt

	The following block decrypts, then encrypts again, and then decrypts
	the last encrypted data in four passes.


*/
	byte_block = NUM_BYTES / 2;

	memcpy(init_vectors, fl_init_vectors, AES_BLOCK_SIZE);

	for(i=0; i < NUM_BYTES; i += byte_block)
	{
		aes_cfb_decrypt(data_in + i, init_vectors, byte_block);
	}

	// Restore startup iV
	memcpy(init_vectors, fl_init_vectors, AES_BLOCK_SIZE);

	for(i=0; i < NUM_BYTES; i += byte_block)
	{
		aes_cfb_encrypt(data_in + i, init_vectors, byte_block);
	}

	// Restore startup iV
	memcpy(init_vectors, fl_init_vectors, AES_BLOCK_SIZE);

	for(i=0; i < NUM_BYTES; i += byte_block)
	{
		aes_cfb_decrypt(data_in + i, init_vectors, byte_block);
	}

#endif



#ifdef TEST
	printf("\nDecrypted Packet: \n\n");
	j=0;
	for(i=0;i<NUM_BYTES;i++){
		printf("%02x ", data_in[i]);
		if(++j==16){
			j=0;
			printf("\n");
		}
	}

	printf("\nDecrypted Packet Characters: \n\n");
	j=0;
	for(i=0;i<NUM_BYTES;i++)
	{
		if((data_in[i] >= 0x20) && (data_in[i] <= 0x7F))
		{
			printf("%c", data_in[i]);
		}
		else
		{
			printf("*");
		}
		if(++j==64){
			j=0;
			printf("\n");
		}
	}


	for(;;){
 		delay_ms(1000);
    }
#endif
}